use crate::span::Span;

/// Kinds of errors that may occur during lexical analysis.
#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub enum LexerErrorKind {
    /// An invalid escape sequence was encountered.
    InvalidEscape(String),
    /// An invalid numeric literal was encountered.
    InvalidNumber(String),
    /// An invalid string literal was encountered.
    InvalidString,
    /// An unclosed character literal was encountered.
    UnclosedChar,
    /// And unclosed string literal was encountered.
    UnclosedString,
}

/// An error which occurred during lexical analysis.
///
/// `LexerError`s contain the kind of error which occurred, as well as a [Span]
/// specifying the [Source] and [Location] of the error.
///
/// [Source]: crate::span::Source
/// [Location]: crate::span::Location
#[derive(Debug, Clone, PartialEq, Hash)]
pub struct LexerError {
    /// The kind of error encountered.
    pub kind: LexerErrorKind,
    /// The span in which the error occurred.
    pub span: Span,
}

impl LexerError {
    /// Construct a new instance of `LexerError`.
    #[must_use]
    pub const fn new(kind: LexerErrorKind, span: Span) -> Self {
        Self { kind, span }
    }
}

impl std::error::Error for LexerError {}

#[cfg(not(tarpaulin_include))]
impl std::fmt::Display for LexerError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use LexerErrorKind::*;

        match &self.kind {
            InvalidEscape(c) => write!(f, "Invalid escape sequence '\\{c}'"),
            InvalidNumber(n) => write!(f, "Invalid numeric literal `{n}`"),
            InvalidString => write!(f, "Invalid string literal"),
            UnclosedChar => write!(f, "Unclosed character literal"),
            UnclosedString => write!(f, "Unclosed string literal"),
        }
    }
}
