use super::Symbol;
use crate::span::Span;

/// Possible kinds of a [Token].
#[derive(Debug, Clone, PartialEq)]
pub(crate) enum TokenKind {
    /// Block comment, e.g. `#| ... |#`
    BlockComment(String),
    /// Line comment, e.g. `; ...`
    LineComment(String),

    /// Opening parenthesis, e.g. `(`
    OpenParen,
    /// Closing parenthesis, e.g. `)`
    CloseParen,
    /// Opening brace, e.g. `{`
    OpenBrace,
    /// Closing brace, e.g. `}`
    CloseBrace,
    /// Opening bracket, e.g. `[`
    OpenBracket,
    /// Closing bracket, e.g. `]`
    CloseBracket,
    /// Opening hash-brace, e.g. `#{`
    OpenHashBrace,

    /// Boolean, e.g. `true`, `false`
    Bool(bool),
    /// Character, e.g. `'c'`, `'\n'`
    Char(String),
    /// Floating-point number, e.g. `-1.0`, `2.0`, `+0.003`
    Float(f64),
    /// Integer, e.g. `0`, `-1`, `+200`
    Integer(i64),
    /// Keyword, e.g. `:baz`
    Keyword(Symbol),
    /// String, e.g. `"foo bar"`
    String(String),
    /// Symbol, e.g. `qux`, `+`
    Symbol(Symbol),
    /// Nil, e.g. `nil`
    Nil,
}

/// A token encountered during lexical analysis.
///
/// `Token`s contain the kind of token which was found, as well as a [Span]
/// specifying the [Source] and [Location] of the token.
///
/// [Source]: crate::span::Source
/// [Location]: crate::span::Location
#[derive(Debug, Clone, PartialEq)]
pub(crate) struct Token {
    /// The kind of token.
    pub(crate) kind: TokenKind,
    /// The span in which the token occurs.
    pub(crate) span: Span,
}

impl Token {
    /// Construct a new instance of `Token`.
    #[must_use]
    pub(crate) const fn new(kind: TokenKind, span: Span) -> Self {
        Self { kind, span }
    }
}
