use super::error::ParserError;
use crate::{
    lexer::{Symbol, Token, TokenKind},
    span::Span,
};

/// Abstract Syntax Tree (AST).
#[derive(Debug, Default, Clone, PartialEq)]
pub(crate) struct Ast {
    root: Vec<Node>,
}

impl Ast {
    /// Construct a new instance of `Ast`.
    #[must_use]
    pub(crate) fn new(root: Vec<Node>) -> Self {
        Self { root }
    }
}

impl From<Vec<Node>> for Ast {
    fn from(root: Vec<Node>) -> Self {
        Self { root }
    }
}

#[cfg(not(tarpaulin_include))]
impl std::fmt::Display for Ast {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for node in &self.root {
            writeln!(f, "{node}")?;
        }

        Ok(())
    }
}

/// A node in the Abstract Syntax Tree (AST).
///
/// `Nodes`s contain the kind of node which was found, as well as a [Span]
/// specifying the [Source] and [Location] of the node.
///
/// [Source]: crate::span::Source
/// [Location]: crate::span::Location
#[derive(Debug, Clone, PartialEq)]
pub(crate) struct Node {
    /// The kind of node.
    pub kind: Expr,
    /// The span in which the node occurs.
    pub span: Span,
}

impl Node {
    /// Construct a new instance of `Node`.
    #[must_use]
    pub(crate) fn new(kind: Expr, span: Span) -> Self {
        Self { kind, span }
    }

    /// Push a child node onto a list node.
    pub(crate) fn push_node(&mut self, child: Self) -> Result<(), ParserError> {
        match &mut self.kind {
            Expr::List(vec) | Expr::Map(vec) | Expr::Set(vec) | Expr::Vector(vec) => {
                vec.push(child);
            }
            _ => unimplemented!(),
        }

        Ok(())
    }

    #[cfg(not(tarpaulin_include))]
    #[must_use]
    fn display(&self, indent: usize) -> String {
        let mut text = format!(
            "{}{}@{}..{}\n",
            "  ".repeat(indent),
            self.kind,
            self.span.bytes().start,
            self.span.bytes().end
        );

        match &self.kind {
            Expr::Atom(_) => {}
            Expr::List(vec) | Expr::Map(vec) | Expr::Set(vec) | Expr::Vector(vec) => {
                for node in vec {
                    text.push_str(&format!("{}\n", node.display(indent + 1)));
                }
            }
        }

        text.trim_end().to_string()
    }
}

impl TryFrom<Token> for Node {
    type Error = ParserError;

    fn try_from(token: Token) -> Result<Self, Self::Error> {
        let span = token.span.clone();
        let kind = Expr::try_from(token)?;

        Ok(Self::new(kind, span))
    }
}

#[cfg(not(tarpaulin_include))]
impl std::fmt::Display for Node {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.display(0))
    }
}

/// An atomic value.
#[derive(Debug, Clone, PartialEq)]
pub(crate) enum Atom {
    /// Boolean, e.g. `true`, `false`
    Bool(bool),
    /// Character, e.g. `'c'`, `'\n'`
    Char(String),
    /// Floating-point number, e.g. `-1.0`, `2.0`, `+0.003`
    Float(f64),
    /// Integer, e.g. `0`, `-1`, `+200`
    Integer(i64),
    /// Keyword, e.g. `:baz`
    Keyword(Symbol),
    /// String, e.g. `"foo bar"`
    String(String),
    /// Symbol, e.g. `qux`, `+`
    Symbol(Symbol),
    /// Nil, e.g. `nil`
    Nil,
}

#[cfg(not(tarpaulin_include))]
impl std::fmt::Display for Atom {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use Atom::*;

        match self {
            Bool(_) => write!(f, "BOOL"),
            Char(_) => write!(f, "CHAR"),
            Float(_) => write!(f, "FLOAT"),
            Integer(_) => write!(f, "INTEGER"),
            Keyword(_) => write!(f, "KEYWORD"),
            String(_) => write!(f, "STRING"),
            Symbol(_) => write!(f, "SYMBOL"),
            Nil => write!(f, "NIL"),
        }
    }
}

/// An expression.
#[derive(Debug, Clone, PartialEq)]
pub(crate) enum Expr {
    /// An atomic value.
    Atom(Atom),
    /// A list of nodes.
    List(Vec<Node>),
    /// A map of nodes.
    Map(Vec<Node>),
    /// A set of nodes.
    Set(Vec<Node>),
    /// A vector of nodes.
    Vector(Vec<Node>),
}

impl Expr {
    /// Which closing delimiter is associated with the expression kind?
    pub(crate) fn closing_delimiter(&self) -> Option<TokenKind> {
        match self {
            Expr::List(_) => Some(TokenKind::CloseParen),
            Expr::Map(_) | Expr::Set(_) => Some(TokenKind::CloseBrace),
            Expr::Vector(_) => Some(TokenKind::CloseBracket),
            _ => None,
        }
    }
}

impl From<Atom> for Expr {
    fn from(atom: Atom) -> Self {
        Self::Atom(atom)
    }
}

impl TryFrom<Token> for Expr {
    type Error = ParserError;

    fn try_from(token: Token) -> Result<Self, Self::Error> {
        let kind = match token.kind {
            TokenKind::Bool(b) => Atom::Bool(b),
            TokenKind::Char(c) => Atom::Char(c),
            TokenKind::Float(n) => Atom::Float(n),
            TokenKind::Integer(n) => Atom::Integer(n),
            TokenKind::Keyword(k) => Atom::Keyword(k),
            TokenKind::String(s) => Atom::String(s),
            TokenKind::Symbol(s) => Atom::Symbol(s),
            TokenKind::Nil => Atom::Nil,
            _ => unimplemented!(),
        };

        Ok(kind.into())
    }
}

#[cfg(not(tarpaulin_include))]
impl std::fmt::Display for Expr {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        use Expr::*;

        match self {
            Atom(atom) => write!(f, "{atom}"),
            List(_) => write!(f, "LIST"),
            Map(_) => write!(f, "MAP"),
            Set(_) => write!(f, "SET"),
            Vector(_) => write!(f, "VECTOR"),
        }
    }
}
