pub(crate) use self::{ast::Ast, error::ParserError};
use self::{
    ast::{Expr, Node},
    error::ParserErrorKind,
};
use crate::{
    lexer::{Lexer, TokenKind},
    span::Span,
};

mod ast;
mod error;

pub(crate) struct Parser<'parser> {
    lexer: Lexer<'parser>,
    parents: Vec<Node>,
    current: Node,
}

impl<'parser> Parser<'parser> {
    /// Create a new parser instance from a string.
    #[must_use]
    pub(crate) fn new(input: &'parser str) -> Self {
        let lexer = Lexer::new(input);
        let current = Node::new(Expr::List(Vec::new()), lexer.span());

        Self {
            lexer,
            parents: Vec::new(),
            current,
        }
    }

    /// Set the name of the lexer's source.
    pub(crate) fn set_name(&mut self, name: String) {
        self.lexer.set_name(name);
    }

    /// Produce an Abstract Syntax Tree (AST) from the source input.
    pub(crate) fn parse(mut self) -> Result<Ast, ParserError> {
        // This parser is actually quite simple!Recursively parse expressions until we
        // run out of tokens, or an error occurs:
        while !self.lexer.eof() {
            if let Some(node) = self.expr()? {
                self.current.push_node(node)?;
            }
        }

        // When we reach the end of input, there should be no remaining parent nodes; if
        // there are, that means that there is a missing closing delimiter somewhere:
        if !self.parents.is_empty() {
            return Err(ParserError::new(
                ParserErrorKind::UnclosedSequence,
                self.current.span,
            ));
        }

        // Since we created an initial `Expr::List` node to hold the parsed contents
        // (i.e. so that we had something to push nodes to), we can now rip out its guts
        // and return the newly constructed AST:
        if let Expr::List(root) = self.current.kind {
            Ok(Ast::new(root))
        } else {
            unreachable!() // TODO: Is this really true? It should be... right?
        }
    }

    fn expr(&mut self) -> Result<Option<Node>, ParserError> {
        if let Some(token) = self.lexer.read()? {
            match token.kind {
                // Comments are simply ignored by the parser:
                TokenKind::BlockComment(_) | TokenKind::LineComment(_) => Ok(None),

                // Any valid opening delimiters begins a new sequence:
                TokenKind::OpenParen => self.begin_sequence(Expr::List, token.span),
                TokenKind::OpenHashBrace => self.begin_sequence(Expr::Map, token.span),
                TokenKind::OpenBrace => self.begin_sequence(Expr::Set, token.span),
                TokenKind::OpenBracket => self.begin_sequence(Expr::Vector, token.span),

                // Any valid closing delimiters end the current sequence:
                kind
                @ (TokenKind::CloseParen | TokenKind::CloseBrace | TokenKind::CloseBracket) => {
                    self.end_sequence(kind, token.span)
                }

                // Atoms are pushed to the current sequence:
                _ => {
                    let node = Node::try_from(token)?;
                    let span = node.span.clone();

                    self.current.push_node(node)?;
                    self.current.span.extend(&span);

                    Ok(None)
                }
            }
        } else {
            Err(ParserError::new(
                ParserErrorKind::UnexpectedEof,
                self.lexer.span(),
            ))
        }
    }

    fn begin_sequence(
        &mut self,
        init: impl FnOnce(Vec<Node>) -> Expr,
        span: Span,
    ) -> Result<Option<Node>, ParserError> {
        self.current.span.extend(&span);
        self.parents.push(self.current.clone());

        self.current = Node::new(init(Vec::new()), span.clone());

        Ok(None)
    }

    fn end_sequence(&mut self, kind: TokenKind, span: Span) -> Result<Option<Node>, ParserError> {
        // We will ultimately return the current expression, so clone it and update its
        // span first:
        let mut current = self.current.clone();
        current.span.extend(&span);

        // Update the parser's current node to the previous parent, or return an error
        // if no parents exist:
        self.current = self.parents.pop().ok_or_else(|| {
            ParserError::new(ParserErrorKind::UnexpectedClosingDelimiter, span.clone())
        })?;

        // Ensure that the appropriate closing delimiter was found for our current node:
        if current.kind.closing_delimiter() != Some(kind) {
            return Err(ParserError::new(
                ParserErrorKind::UnmatchedClosingDelimiter,
                span,
            ));
        }

        // For maps, ensure that each key has a corresponding value:
        match current.kind {
            Expr::Map(ref vec) if vec.len() % 2 != 0 => {
                return Err(ParserError::new(ParserErrorKind::MissingValueInMap, span));
            }
            _ => {}
        }

        // Finally, return the current node so it can be added to the AST:
        Ok(Some(current))
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{
        lexer::{LexerError, LexerErrorKind, Symbol},
        parser::ast::Atom,
    };

    macro_rules! test {
        ( $name:ident: $input:literal, $src:ident => $ast:expr ) => {
            #[test]
            fn $name() {
                let parser = Parser::new($input);
                let $src = parser.lexer.source();
                assert_eq!(parser.parse(), $ast);
            }
        };
    }

    test!(empty: "", _src => Ok(Ast::default()));

    test!(error_invalid_number: "(+ 1.2.3)", src => Err(ParserError::new(
        ParserErrorKind::Lexer(LexerError::new(
            LexerErrorKind::InvalidNumber("1.2.3".into()),
            Span::new(3..8, src.clone())
        )),
        Span::new(3..8, src)
    )));

    test!(list: "(+ 1 2) ; sneaky comment :)", src => Ok(Ast::from(vec![
        Node::new(
            Expr::List(vec![
                Node::new(Atom::Symbol(Symbol::from("+")).into(), Span::new(1..2, src.clone())),
                Node::new(Atom::Integer(1).into(), Span::new(3..4, src.clone())),
                Node::new(Atom::Integer(2).into(), Span::new(5..6, src.clone())),
            ]),
            Span::new(0..7, src)
        )
    ])));

    test!(error_list_unmatched_bracket: "(]", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedClosingDelimiter,
        Span::new(1..2, src),
    )));

    test!(error_list_missing_close_paren: "(true", src => Err(ParserError::new(
        ParserErrorKind::UnclosedSequence,
        Span::new(0..5, src),
    )));

    test!(error_list_unexpected_close_paren: ")", src => Err(ParserError::new(
        ParserErrorKind::UnexpectedClosingDelimiter,
        Span::new(0..1, src)
    )));

    test!(map: "#{:a 0.0 :b 1.0}", src => Ok(Ast::from(vec![
        Node::new(
            Expr::Map(vec![
                Node::new(Atom::Keyword(Symbol::from("a")).into(), Span::new(2..4, src.clone())),
                Node::new(Atom::Float(0.0).into(), Span::new(5..8, src.clone())),
                Node::new(Atom::Keyword(Symbol::from("b")).into(), Span::new(9..11, src.clone())),
                Node::new(Atom::Float(1.0).into(), Span::new(12..15, src.clone())),
            ]),
            Span::new(0..16, src)
        )
    ])));

    test!(error_map_missing_value: "#{:x}", src => Err(ParserError::new(
        ParserErrorKind::MissingValueInMap,
        Span::new(4..5, src.clone())
    )));

    test!(error_map_unmatched_bracket: "#{)", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedClosingDelimiter,
        Span::new(2..3, src),
    )));

    test!(error_map_missing_close_brace: "#{", src => Err(ParserError::new(
        ParserErrorKind::UnclosedSequence,
        Span::new(0..2, src),
    )));

    test!(error_map_set_unexpected_close_brace: "}", src => Err(ParserError::new(
        ParserErrorKind::UnexpectedClosingDelimiter,
        Span::new(0..1, src)
    )));

    test!(set: "{{} nil}", src => Ok(Ast::from(vec![
        Node::new(
            Expr::Set(vec![
                Node::new(Expr::Set(vec![]), Span::new(1..3, src.clone())),
                Node::new(Expr::Atom(Atom::Nil), Span::new(4..7, src.clone())),
            ]),
            Span::new(0..8, src),
        )
    ])));

    test!(error_set_unmatched_bracket: "{]", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedClosingDelimiter,
        Span::new(1..2, src),
    )));

    test!(error_set_missing_close_brace: "{", src => Err(ParserError::new(
        ParserErrorKind::UnclosedSequence,
        Span::new(0..1, src),
    )));

    test!(vector: "['a' 'b' 'c']", src => Ok(Ast::from(vec![
        Node::new(
            Expr::Vector(vec![
                Node::new(Expr::Atom(Atom::Char("a".into())), Span::new(1..4, src.clone())),
                Node::new(Expr::Atom(Atom::Char("b".into())), Span::new(5..8, src.clone())),
                Node::new(Expr::Atom(Atom::Char("c".into())), Span::new(9..12, src.clone())),
            ]),
            Span::new(0..13, src),
        )
    ])));

    test!(error_vector_unmatched_bracket: "[}", src => Err(ParserError::new(
        ParserErrorKind::UnmatchedClosingDelimiter,
        Span::new(1..2, src),
    )));

    test!(error_vector_missing_close_bracket: "[", src => Err(ParserError::new(
        ParserErrorKind::UnclosedSequence,
        Span::new(0..1, src),
    )));

    test!(error_vector_unexpected_close_bracket: "]", src => Err(ParserError::new(
        ParserErrorKind::UnexpectedClosingDelimiter,
        Span::new(0..1, src)
    )));

    test!(multiple_expressions: "(/ 6 3 (+ 1 2)) (* 2 5)\n(- 10 5)", src => Ok(Ast::from(vec![
        Node::new(
            Expr::List(vec![
                Node::new(Atom::Symbol(Symbol::from("/")).into(), Span::new(1..2, src.clone())),
                Node::new(Atom::Integer(6).into(), Span::new(3..4, src.clone())),
                Node::new(Atom::Integer(3).into(), Span::new(5..6, src.clone())),
                Node::new(
                    Expr::List(vec![
                        Node::new(Atom::Symbol(Symbol::from("+")).into(), Span::new(8..9, src.clone())),
                        Node::new(Atom::Integer(1).into(), Span::new(10..11, src.clone())),
                        Node::new(Atom::Integer(2).into(), Span::new(12..13, src.clone())),
                    ]),
                    Span::new(7..14, src.clone())
                ),
            ]),
            Span::new(0..15, src.clone())
        ),
        Node::new(
            Expr::List(vec![
                Node::new(Atom::Symbol(Symbol::from("*")).into(), Span::new(17..18, src.clone())),
                Node::new(Atom::Integer(2).into(), Span::new(19..20, src.clone())),
                Node::new(Atom::Integer(5).into(), Span::new(21..22, src.clone())),
            ]),
            Span::new(16..23, src.clone())
        ),
        Node::new(
            Expr::List(vec![
                Node::new(Atom::Symbol(Symbol::from("-")).into(), Span::new(25..26, src.clone())),
                Node::new(Atom::Integer(10).into(), Span::new(27..29, src.clone())),
                Node::new(Atom::Integer(5).into(), Span::new(30..31, src.clone())),
            ]),
            Span::new(24..32, src)
        ),
    ])));

    test!(function_application: "(join \"foo\" \"bar\")", src => Ok(Ast::from(vec![Node::new(
        Expr::List(vec![
            Node::new(Atom::Symbol(Symbol::from("join")).into(), Span::new(1..5, src.clone())),
            Node::new(Atom::String("foo".into()).into(), Span::new(6..11, src.clone())),
            Node::new(Atom::String("bar".into()).into(), Span::new(12..17, src.clone())),
        ]),
        Span::new(0..18, src)
    )])));
}
