use std::{cmp::Ordering, iter, ops::Range, sync::Arc};

/// A location within some source text.
#[derive(Debug, Default, Clone, Copy, PartialEq, Eq, Hash)]
pub struct Location {
    line: usize,
    column: usize,
}

impl Location {
    /// Construct a new instance of `Location`.
    #[must_use]
    pub(crate) const fn new(line: usize, column: usize) -> Self {
        Self { line, column }
    }
}

impl PartialOrd for Location {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        match self.line.partial_cmp(&other.line) {
            Some(Ordering::Equal) => self.column.partial_cmp(&other.column),
            ord => ord,
        }
    }
}

/// Some (optionally named) source text.
#[derive(Debug, Default, Clone, PartialEq, Eq, Hash)]
pub struct Source {
    name: Option<String>,
    contents: String,
    lines: Vec<usize>,
}

impl Source {
    /// Construct a new instance of `Source`.
    #[must_use]
    pub(crate) fn new(name: Option<String>, contents: String) -> Self {
        let lines = contents
            .match_indices('\n')
            .map(|(i, _)| i)
            .chain(iter::once(contents.len()))
            .collect();

        Self {
            name,
            contents,
            lines,
        }
    }

    /// Get the name of the source.
    #[must_use]
    pub(crate) fn name(&self) -> Option<&str> {
        self.name.as_deref()
    }

    /// Set the name of the source.
    pub(crate) fn set_name(&mut self, name: String) {
        self.name = Some(name);
    }

    /// Get the [Location] of the specified byte in the source.
    #[must_use]
    pub(crate) fn location(&self, byte: usize) -> Location {
        let line = self.lines.partition_point(|&x| x < byte);
        let start = line.checked_sub(1).map_or(0, |n| self.lines[n] + 1);
        let column = self.contents[start..byte].chars().count();

        Location::new(line, column)
    }

    /// Get the full contents of the source.
    #[must_use]
    pub(crate) fn contents(&self) -> &str {
        &self.contents
    }

    /// Get the specified line from the source.
    #[must_use]
    pub(crate) fn get_line(&self, line: usize) -> &str {
        let end = self.lines[line];
        let start = line.checked_sub(1).map_or(0, |n| self.lines[n] + 1);

        &self.contents[start..end]
    }
}

/// A contiguous sequence of bytes within some source.
#[derive(Debug, Default, Clone, Eq)]
pub struct Span {
    bytes: Range<usize>,
    source: Arc<Source>,
}

impl Span {
    /// Construct a new instance of `Span`.
    #[must_use]
    pub(crate) fn new(bytes: Range<usize>, source: Arc<Source>) -> Self {
        Self { bytes, source }
    }

    /// Join two spans, creating a new span.
    #[must_use]
    pub(crate) fn join(self, other: &Self) -> Self {
        debug_assert!(self.same_source(other));

        Self::new(self.bytes.start..other.bytes.end, self.source)
    }

    /// Extend one span to include another.
    pub(crate) fn extend(&mut self, other: &Self) {
        debug_assert!(self.same_source(other));
        self.bytes.end = other.bytes.end;
    }

    /// The start location of a span within some source.
    #[must_use]
    pub(crate) fn location(&self) -> Location {
        self.source.location(self.bytes.start)
    }

    /// The end location of a span within some source.
    #[must_use]
    pub(crate) fn end_location(&self) -> Location {
        self.source.location(self.bytes.end)
    }

    /// Do two spans share the same source?
    #[must_use]
    pub(crate) fn same_source(&self, other: &Self) -> bool {
        Arc::ptr_eq(&self.source, &other.source)
    }

    #[must_use]
    pub(crate) fn bytes(&self) -> &Range<usize> {
        &self.bytes
    }
}

impl PartialEq for Span {
    fn eq(&self, other: &Self) -> bool {
        self.same_source(other) && self.bytes == other.bytes
    }
}

impl std::hash::Hash for Span {
    fn hash<H>(&self, state: &mut H)
    where
        H: std::hash::Hasher,
    {
        self.bytes.hash(state);
        self.source.hash(state);
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn location_partial_ord() {
        assert!(Location::new(1, 1) < Location::new(1, 2));
        assert!(Location::new(1, 10) < Location::new(2, 1));
        assert!(Location::new(5, 5) == Location::new(5, 5));
        assert!(Location::new(10, 1) > Location::new(9, 99));
    }

    #[test]
    fn source_get_set_name() {
        let mut src = Source::new(None, "".into());
        assert!(src.name().is_none());

        src.set_name("foo".into());
        assert!(src.name() == Some("foo"));
    }

    #[test]
    fn source_location() {
        let source = Source::new(None, "foo\nbar\nbaz".into());
        assert_eq!(source.location(0), Location::new(0, 0));
        assert_eq!(source.location(5), Location::new(1, 1));
        assert_eq!(source.location(10), Location::new(2, 2));
    }

    #[test]
    fn source_contents() {
        let contents = String::from("xxx");
        let source = Source::new(None, contents.clone());
        assert_eq!(source.contents(), &contents);
    }

    #[test]
    fn source_get_line() {
        let source = Source::new(None, "line 1\nline 2\nline 3\n".into());
        assert_eq!(source.get_line(0), "line 1");
        assert_eq!(source.get_line(1), "line 2");
        assert_eq!(source.get_line(2), "line 3");
    }

    #[test]
    fn span_partial_eq() {
        let source = Arc::new(Source::new(None, String::new()));

        let a = Span::new(0..0, source.clone());
        assert_eq!(a, a);

        let b = Span::new(1..10, source.clone());
        assert_ne!(a, b);

        let source2 = Arc::new(Source::new(None, String::from("foo")));
        let c = Span::new(0..0, source2.clone());
        assert_ne!(a, c);
    }

    #[test]
    fn span_start_end_location() {
        let source = Arc::new(Source::new(None, "foo\nbar\nbaz".into()));
        let span = Span::new(2..9, source);

        let start = span.location();
        assert_eq!(start.line, 0);
        assert_eq!(start.column, 2);

        let end = span.end_location();
        assert_eq!(end.line, 2);
        assert_eq!(end.column, 1);
    }
}
